"""This module provides BeamSampling class for generating completions using beam search sampling."""

from math import exp, log2
from random import random
from typing import Any, Callable, Iterable, Optional, Tuple

from sign_language_translator.models.language_models.abstract_language_model import (
    LanguageModel,
)
from sign_language_translator.utils import sample_one_index


class BeamSampling:
    """BeamSampling class for generating completions using beam search sampling.

    Args:
        model (LanguageModel): The language model used for generating completions.
        beam_width (int, optional): The beam width for beam search. Defaults to 3.
        start_of_sequence_token (str, optional): The start of sequence token. Defaults to "[".
        end_of_sequence_token (str, optional): The end of sequence token. Defaults to "]".
        max_length (int, optional): The maximum length of the generated completions. Defaults to 33.
        scoring_function (Callable[[Iterable, float], float], optional):
            The scoring function used to score the completions. It should accept the generated
            sequence and its overall log probability as arguments. Defaults to a linear function.

        return_log_of_probability (bool, optional):
            A flag indicating whether to return the probability of the completions
            or log2 of it. Defaults to True.
    """

    def __init__(
        self,
        model: LanguageModel,
        beam_width: int = 3,
        start_of_sequence_token="[",
        end_of_sequence_token="]",
        max_length: int = 37,
        scoring_function: Callable[[Iterable, float], float] = (
            lambda seq, log_prob: 10.0 + log_prob / len(seq)  # type: ignore
        ),
        return_log_of_probability: bool = True,
    ) -> None:
        self.model = model
        self.start_of_sequence_token = start_of_sequence_token
        self.end_of_sequence_token = end_of_sequence_token
        self.beam_width = beam_width
        self.max_length = max_length
        self.scoring_function = scoring_function
        self.return_log_of_probability = return_log_of_probability

    def __call__(self, context: Optional[Iterable] = None) -> Iterable:
        return self.complete(initial_context=context)

    def complete(
        self,
        initial_context: Optional[Iterable] = None,
        append_func: Callable[[Any, Any], Any] = lambda context, token: (
            context + [token]
            if isinstance(context, list)
            else context + (token,)
            if isinstance(context, tuple)
            else context + token
        ),
    ) -> Tuple[Iterable, float]:
        """Generate completions based on the given initial context.

        Args:
            initial_context (Iterable | None, optional): The initial context for completion generation. Defaults to None.
            append_func (Callable[[Any, Any], Any], optional): a function that can append the generated next token to provided context.
                Defaults to a lambda function that can append to list, tuple & str.

        Returns:
            Tuple[Iterable, float]: One generated completion and its score.
        """

        if initial_context is None:
            initial_context = [self.start_of_sequence_token]

        branches = [(initial_context, 0.0)]

        for _ in range(self.max_length):
            n_branches = round(self.beam_width + random() * 0.8 - 0.4)

            # Expand
            new_branches = []
            for context_, score_ in branches:
                if (  # sequence completed
                    context_[-1] == self.end_of_sequence_token  # type: ignore
                    or len(context_) >= self.max_length  # type: ignore
                ):
                    # just as it was. no change.
                    new_branches.append((context_, score_))
                    continue

                # append next tokens
                for _ in range(n_branches):
                    next_token, prob = self.model.next(context_)
                    if next_token == self.model.unknown_token:
                        score = score_
                        next_context = context_
                    else:
                        next_context = append_func(context_, next_token)
                        score = score_ + log2(prob)

                    # no repeats; for diversity.
                    if (next_context, score) not in new_branches:
                        new_branches.append((next_context, score))

            # Prune
            new_branches = sorted(
                new_branches,
                key=lambda item: self.scoring_function(*item),
                reverse=True,
            )[:n_branches]

            if branches == new_branches:
                # no branch has grown further: stop
                break

            branches = new_branches

        # softmax: turn each branch's scores into a probability distribution
        weights = [self.scoring_function(seq, log_prob) for seq, log_prob in branches]
        weights = [exp(w) for w in weights]
        weights = [w / sum(weights) for w in weights]

        # select one branch
        selected_completion, score = branches[sample_one_index(weights)]

        # reformat score
        if not self.return_log_of_probability:
            score = 2**score  # math.exp2(score)

        return selected_completion, score



from sign_language_translator.models.text_embedding.text_embedding_model import (
    TextEmbeddingModel,
)
from sign_language_translator.models.text_embedding.vector_lookup_model import (
    VectorLookupModel,
)

__all__ = [
    "TextEmbeddingModel",
    "VectorLookupModel",
]



"""
Module for VectorLookupModel, a class extending TextEmbeddingModel,
which finds the pretrained embedding of text via hash tables.

Classes:
    - VectorLookupModel: A text embedding model that maps tokens to vectors.
"""

from collections import Counter
from os.path import basename
from typing import Callable, Iterable, List, Optional, Tuple
from zipfile import ZIP_DEFLATED, ZipFile

import numpy as np
import torch

from sign_language_translator.models.text_embedding.text_embedding_model import (
    TextEmbeddingModel,
)


class VectorLookupModel(TextEmbeddingModel):
    """
    VectorLookupModel class extends TextEmbeddingModel to provide text embedding based on pre-defined token vectors.

    Attributes:
        - index_to_token (List[str]): A list containing tokens in the same order as the vectors.
        - known_tokens (frozenset): A frozenset containing unique known tokens.
        - token_to_index (Dict[str, int]): A dictionary mapping tokens to their corresponding indices.
        - vectors (torch.Tensor): A 2D tensor representing the token vectors.

    Methods:
        - update(self, tokens: List[str], vectors: torch.Tensor) -> None: Updates existing tokens & hash-table with new vectors.
        - embed(self, text: str, pre_normalize=False, post_normalize=False,
                tokenizer: Callable[[str], Iterable[str]] = lambda x: x.split()) -> torch.Tensor:
          Returns the pretrained embedding vector for a token or average embedding of sub tokens.
        - __getitem__(self, token: str) -> torch.Tensor: Returns the vector for a specific token.
        - save(self, path: str): Saves the model state (tokens & vectors) to a file.
        - load(cls, path: str): Loads a saved model state (tokens & vectors) from a file.

    Usage:
        ```python
        from sign_language_translator.models import VectorLookupModel
        import torch

        tokens = ["example", "text"]
        vectors = torch.tensor([[1, 2, 3], [4, 5, 6]])
        model = VectorLookupModel(tokens, vectors)

        embedding = model.embed("example text")  # [2.5, 3.5, 4.5]

        model.update(["hello"], torch.tensor([[7, 8, 9]]))

        model.save("model.pt")
        loaded_model = VectorLookupModel.load("model.pt")
        ```
    """

    def __init__(
        self,
        tokens: List[str],
        vectors: torch.Tensor,
        alignment_matrix: Optional[torch.Tensor] = None,
        description: str = "",
    ):
        """
        Initializes the VectorLookupModel.

        Args:
            tokens (List[str]): A list of tokens.
            vectors (torch.Tensor): A (2D) tensor of vectors in the same order as `tokens`.
        """

        self.__validate_init_args(tokens, vectors, alignment_matrix)
        self.index_to_token = tokens
        self.known_tokens = frozenset(tokens)
        self.token_to_index = {token: index for index, token in enumerate(tokens)}
        self.vectors = vectors
        self.alignment_matrix = alignment_matrix
        self.description = description

        self._normalized_vectors = None
        self._tokens_array = None

    def update(self, tokens: List[str], vectors: torch.Tensor) -> None:
        """
        Update the vector lookup model with new tokens and their corresponding vectors.

        Args:
            tokens (List[str]): The list of new tokens to be added or updated.
            vectors (torch.Tensor): The tensor of corresponding vectors for the new tokens.
            alignment_matrix (Optional[torch.Tensor], optional): A 2D Tensor to transform the final vectors. (e.g. some orthogonal matrix can be used to align the word vector to an embedding for some other language or model). Defaults to None.
            description (str, optional): A description of the model. Defaults to "".

        Raises:
            ValueError: If the dimensions of the new vectors do not match the dimensions of the existing vectors.
        """

        # validation
        self.__validate_init_args(tokens, vectors, None)
        if (stored_dims := self.vectors.shape[1:]) != (new_dims := vectors.shape[1:]):
            raise ValueError(f"Dimension mismatch: {stored_dims = }, {new_dims = }.")

        # update previously existing tokens
        _indexes = [
            (self.token_to_index[t], i)
            for i, t in enumerate(tokens)
            if t in self.known_tokens
        ]
        stored_indexes, new_indexes = list(zip(*_indexes))
        self.vectors[stored_indexes] = vectors[new_indexes]

        # filter out new tokens
        remaining_mask = torch.ones(len(tokens), dtype=torch.bool)
        remaining_mask[new_indexes] = False
        new_tokens = [tokens[i] for i in torch.arange(len(tokens))[remaining_mask]]
        old_len = len(self.index_to_token)

        # append & update dependent properties
        self.vectors = torch.cat([self.vectors, vectors[remaining_mask]], dim=0)
        self.index_to_token = self.index_to_token + new_tokens
        self.known_tokens = frozenset(self.index_to_token)
        self.token_to_index.update({t: i + old_len for i, t in enumerate(new_tokens)})

        if self._tokens_array is not None:
            self._tokens_array = np.concatenate(
                [self._tokens_array, np.array(new_tokens)]
            )
        if self._normalized_vectors is not None:
            _norms = vectors[remaining_mask].norm(dim=1, keepdim=True)
            _norms[_norms == 0] = 1
            self._normalized_vectors = torch.cat(
                [self._normalized_vectors, vectors[remaining_mask] / _norms]
            )

    # =============== #
    #    embedding    #
    # =============== #

    def embed(
        self,
        text: str,
        pre_normalize=False,
        post_normalize=False,
        align=False,
        tokenizer: Callable[[str], Iterable[str]] = lambda x: x.split(),
    ) -> torch.Tensor:
        """
        Embeds the given text into a vector representation by lookup or averaging pre-computed embeddings.

        Args:
            text (str): The input text to be embedded, (can be in the model vocabulary or be a string of tokens from the model dictionary). If unknown, returns a zero vector.
            pre_normalize (bool, optional): Whether to normalize the vectors of tokens in the text before averaging. Defaults to False.
            post_normalize (bool, optional): Whether to normalize the vector after embedding. Defaults to False.
            align (bool, optional): Whether to transform the final vector using the alignment matrix. Defaults to False.
            tokenizer (Callable[[str], Iterable[str]], optional): A callable function to tokenize the text. Only used if the text is not present in the model vocabulary. Defaults to splitting on whitespace.

        Returns:
            torch.Tensor: The embedded vector representation of the input text.
        """
        # TODO: Handle batches

        # found in the vocabulary
        if text in self.known_tokens:
            vector = self.vectors[self.token_to_index[text]].clone()
            # scale to unit length
            if pre_normalize:
                if norm := vector.norm():
                    vector = vector / norm
        # Average of known tokens
        else:
            # break text into tokens
            tokens = tokenizer(text)
            indexes = [self.token_to_index[t] for t in tokens if t in self.known_tokens]
            if len(indexes) > 0:
                vectors = self.vectors[indexes].clone()
                # scale to unit length
                if pre_normalize:
                    vectors = vectors / vectors.norm(dim=1, keepdim=True)
                # average
                vector = vectors.nanmean(dim=0)
            # unable to embed
            else:
                vector = torch.zeros(self.vectors.shape[1:])

        # scale to unit length
        if post_normalize:
            if norm := vector.norm():
                vector = vector / norm

        if align and self.alignment_matrix is not None:
            vector = vector @ self.alignment_matrix

        return vector

    def __getitem__(self, token: str) -> torch.Tensor:
        """
        Retrieve the embedding vector for a given token. If token is unknown, returns a zero vector.

        Args:
            token (str): The token for which to retrieve the embedding vector.

        Returns:
            torch.Tensor: The embedding vector for the given token.
        """

        return self.embed(token)

    # ============ #
    #    Search    #
    # ============ #

    @property
    def normalized_vectors(self):
        if self._normalized_vectors is None:
            _norms = self.vectors.norm(dim=1, keepdim=True)
            _norms[_norms == 0] = 1
            self._normalized_vectors = self.vectors / _norms

        return self._normalized_vectors

    @property
    def tokens_array(self):
        if self._tokens_array is None:
            self._tokens_array = np.array(self.index_to_token)
        return self._tokens_array

    def similar(
        self, vector: torch.Tensor, k: int = 1
    ) -> Tuple[List[str], List[float]]:
        """
        Find the k most similar tokens to the given vector.

        Args:
            vector (torch.Tensor): The 1D vector for which to find similar tokens.
            k (int, optional): The number of similar tokens to return. Defaults to 1.

        Returns:
            Tuple[List[str], List[float]]: A tuple containing the k most similar tokens and their corresponding cosine similarities.
        """

        # normalize the query vector
        _norm = vector.norm(keepdim=True)
        _norm[_norm == 0] = 1

        # calculate cosine similarities
        similarities = (vector / _norm) @ self.normalized_vectors.T
        top_k_similarities, top_k_indexes = similarities.topk(k)

        # return the top k similar tokens and their similarities
        return (
            self.tokens_array[top_k_indexes.numpy()].tolist(),
            top_k_similarities.tolist(),
        )

    # =============== #
    #    load/save    #
    # =============== #

    def save(self, path: str):
        """
        Serialize the tokens list and corresponding vectors to a file.
        If the path ends with '.zip' the file will be compressed.

        Args:
            path (str): The path to save the model file.
        """

        data = {
            "tokens": self.index_to_token,
            "vectors": self.vectors,
            "alignment": self.alignment_matrix,
            "description": self.description,
        }
        if path.endswith(".zip"):
            with ZipFile(path, "w", ZIP_DEFLATED) as archive:
                with archive.open(basename(path).replace(".zip", ".pt"), "w") as f:
                    torch.save(data, f)
        else:
            torch.save(data, path)

    @classmethod
    def load(cls, path: str):
        """
        Load a VectorLookupModel from a saved checkpoint.
        If the path ends with '.zip' the file will be decompressed.

        Args:
            path (str): The path to the saved checkpoint.

        Returns:
            VectorLookupModel: The loaded VectorLookupModel instance.
        """

        if path.endswith(".zip"):
            with ZipFile(path, "r") as archive:
                with archive.open(basename(path).replace(".zip", ".pt"), "r") as f:
                    checkpoint = torch.load(f)
        else:
            checkpoint = torch.load(path)

        return cls(
            checkpoint["tokens"],
            checkpoint["vectors"],
            alignment_matrix=checkpoint.get("alignment", None),
            description=checkpoint.get("description", ""),
        )

    # ============= #
    #    Helpers    #
    # ============= #

    def __validate_init_args(self, tokens: List[str], vectors: torch.Tensor, alignment):
        if len(tokens) != vectors.shape[0]:
            raise ValueError(f"Size mismatch: {len(tokens) = }, {len(vectors) = }.")
        if vectors.shape[1] == 0:
            raise ValueError("Expected at least one dimension in vectors.")
        if repeated := {t: c for t, c in Counter(tokens).items() if c > 1}:
            raise ValueError(f"Tokens must be unique. Repeated tokens: {repeated}")
        if alignment is not None:
            if alignment.shape[0] != vectors.shape[1]:
                raise ValueError(f"({alignment.shape[0]= }) != ({vectors.shape[1]= })")

    def __len__(self):
        return len(self.index_to_token)



from typing import Iterable, List, Optional, Sequence, Tuple, Type, Union

import numpy as np
from numpy.typing import NDArray
from torch import Tensor
from torch import long as torch_long

__all__ = [
    "ArrayOps",
    "linear_interpolation",
    "adjust_vector_angle",
    "align_vectors",
]


class ArrayOps:
    @staticmethod
    def floor(array: Union[NDArray, Tensor]) -> Union[NDArray, Tensor]:
        if isinstance(array, np.ndarray):
            return np.floor(array)
        elif isinstance(array, Tensor):
            return array.floor()
        else:
            raise TypeError(f"Invalid type for flooring: {type(array)}")

    @staticmethod
    def ceil(array: Union[NDArray, Tensor]) -> Union[NDArray, Tensor]:
        if isinstance(array, np.ndarray):
            return np.ceil(array)
        elif isinstance(array, Tensor):
            return array.ceil()
        else:
            raise TypeError(f"Invalid type for ceiling: {type(array)}")

    @staticmethod
    def take(
        array: Union[NDArray, Tensor], index: Union[NDArray, Tensor, List], dim: int = 0
    ) -> Union[NDArray, Tensor]:
        if isinstance(array, np.ndarray):
            if not isinstance(index, np.ndarray):
                index = np.array(index)
            return np.take(array, index.astype(int), axis=dim)
        elif isinstance(array, Tensor):
            if not isinstance(index, Tensor):
                index = Tensor(index)
            return array.index_select(dim, index.type(torch_long))
        else:
            raise TypeError(f"Invalid type for taking: {type(array)}")

    @staticmethod
    def cast(
        x: Union[NDArray, Tensor, List, Iterable],
        data_type: Type[Union[np.ndarray, Tensor]],
    ) -> Union[NDArray, Tensor]:
        if data_type == np.ndarray:
            return np.array(x)
        elif data_type == Tensor:
            return Tensor(x)
        else:
            raise TypeError(f"Invalid type for array casting: {data_type}")

    @staticmethod
    def norm(
        x: Union[NDArray, Tensor, List, Iterable],
        dim: Optional[int] = None,
        keepdim=False,
    ) -> Union[NDArray, Tensor]:
        """
        Compute the norm of a given array or tensor along a specified dimension.

        Args:
            x (Union[NDArray, Tensor, List, Iterable]): The input array or tensor.
            dim (Optional[int]): The dimension along which to compute the norm. If None, the norm is computed over the entire array or tensor. Default is None.
            keepdim (bool): Whether to keep the dimension of the input array or tensor after computing the norm. Default is False.

        Returns:
            Union[NDArray, Tensor]: The norm of the input array or tensor.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(x, np.ndarray):
            return np.linalg.norm(x, axis=dim, keepdims=keepdim)  # type: ignore
        elif isinstance(x, Tensor):
            return x.norm(dim=dim, keepdim=keepdim)
        else:
            raise TypeError(f"Invalid type for norm: {type(x)}")

    @staticmethod
    def svd(
        x: Union[NDArray, Tensor, List, Iterable]
    ) -> Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor], Union[NDArray, Tensor]]:
        """
        Compute the singular value decomposition of a given array or tensor.

        Args:
            x (Union[NDArray, Tensor, List, Iterable]): The input array or tensor.

        Returns:
            Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor], Union[NDArray, Tensor]]: The (Rotation, coordinate scaling, reflection) matrices of the input array or tensor.

        Raises:
            TypeError: If the input type is not supported.
        """

        if not isinstance(x, (List, Iterable)):
            x = np.array(x)

        if isinstance(x, np.ndarray):
            return np.linalg.svd(x)
        if isinstance(x, Tensor):
            U, S, V = x.svd()
            return U, S, V.T

        raise TypeError(f"Invalid type for svd: {type(x)}")

    @staticmethod
    def top_k(
        x: Union[NDArray, Tensor, List, Iterable], k: int, dim: int = -1, largest=True
    ) -> Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor]]:
        """
        Compute the top k values and their indices along a specified dimension of a given array or tensor.

        Args:
            x (Union[NDArray, Tensor, List, Iterable]): The input array or tensor.
            k (int): The number of top values to return.
            dim (int, optional): The dimension along which to compute the top k values. Default is -1.
            largest (bool, optional): Whether to return the largest or smallest k values. Default is True.

        Returns:
            Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor]]: The top k values and their indices along the specified dimension.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(x, np.ndarray):
            indices = (
                np.argpartition(x, -k, axis=dim)[-k:]
                if largest
                else np.argpartition(x, k, axis=dim)[:k]
            )
            return x.take(indices, axis=dim), indices
        if isinstance(x, Tensor):
            values, indices = x.topk(k, dim=dim, largest=largest, sorted=True)
            return values, indices

        raise TypeError(f"Invalid type for topk: {type(x)}")


def linear_interpolation(
    array: Union[NDArray[np.float64], Tensor, List],
    new_indexes: Optional[Sequence[Union[int, float]]] = None,
    old_x: Optional[Sequence[Union[int, float]]] = None,
    new_x: Optional[Sequence[Union[int, float]]] = None,
    dim: int = 0,
) -> Union[NDArray, Tensor]:
    """
    Perform linear interpolation on a multidimensional array or tensor along a dimension.

    This function essentially connects all consecutive values in a multidimensional array with
    straight lines along a specified dimension, so that intermediate values can be calculated.
    It takes the input array, a set of new indexes or alternatively old & new coordinate axes,
    and a dimension along which to perform interpolation.

    Parameters:
        array (NDArray[np.float64] | Tensor): The input array or tensor to interpolate.
        new_indexes (Sequence[int | float] | None, optional): The new indexes at which to interpolate the data.
            If None, it infers new_indexes from `old_x` and `new_x`.
        old_x (Sequence[int | float] | None, optional): The old coordinate values corresponding to the data in `array`.
            If None, it uses `new_indexes` argument.
        new_x (Sequence[int | float] | None, optional): The new coordinate values corresponding to the .
            If None, it uses `new_indexes` argument.
        dim (int, optional): The dimension along which to perform interpolation. Default is 0.

    Returns:
        NDArray | Tensor: The result of linear interpolation along the specified dimension.

    Raises:
        ValueError: If both or neither of `new_indexes` and `old_x` & `new_x` are provided.
        If `new_indexes` is not 1 dimensional.

    Examples:
        .. code-block: python
            data = np.array([1, 2, 3, 5])
            new_indexes = np.array([1.5, 0.5, 2.5])
            interpolated_data = linear_interpolation(data, new_indexes)
            print(interpolated_data)
            # array([2.5, 1.5, 4. ])

            old_x = [0, 4, 4.5, 5]
            new_x = [0, 1, 2, 2.5, 3, 4, 5]
            interpolated_data = linear_interpolation(data, old_x=old_x, new_x=new_x)
            print(interpolated_data)
            # array([1.   , 1.25 , 1.5  , 1.625, 1.75 , 2.   , 5.   ])

    Note:
        This function supports both NumPy arrays and PyTorch tensors as input.

    """
    if isinstance(array, list):
        array = np.array(array)

    indexes_, dim = __validate_linear_interpolation_args(
        array, new_indexes, old_x, new_x, dim
    )

    # interpolate
    floor_values = ArrayOps.floor(indexes_)
    ceiling_values = ArrayOps.ceil(indexes_)
    fraction = (indexes_ - floor_values).reshape(
        [len(indexes_) if axis == dim else 1 for axis in range(array.ndim)]
    )

    interpolated = ArrayOps.take(array, floor_values, dim) * (1 - fraction)
    interpolated += ArrayOps.take(array, ceiling_values, dim) * fraction

    return interpolated


def __validate_linear_interpolation_args(array, new_indexes, old_x, new_x, dim):
    if (new_indexes is None and (old_x is None or new_x is None)) or (
        new_indexes is not None and (old_x is not None or new_x is not None)
    ):
        raise ValueError(
            "Either `new_indexes` or both `old_x` and `new_x` must be provided."
        )
    if new_indexes is None:
        new_indexes = np.interp(new_x, old_x, np.arange(len(old_x)))  # type: ignore

    if not isinstance(array, (np.ndarray, Tensor)):
        array = np.array(array)
    indexes_ = ArrayOps.cast(new_indexes, type(array))  # type: ignore

    if indexes_.ndim != 1:
        raise ValueError(
            f"Invalid indexes shape: {indexes_.shape}. Must be 1-dimensional."
        )
    if dim >= array.ndim:
        raise ValueError(f"Invalid dim: {dim}. Must be between 0 and {array.ndim-1}.")

    out_of_bounds = (indexes_ < 0) & (indexes_ > array.shape[dim] - 1)
    if out_of_bounds.any():
        raise ValueError(
            f"Invalid indexes for interpolation: {indexes_[out_of_bounds.tolist()]}. "
            f"Must be between >= 0 and <= {array.shape[0]-1}."
        )
    if dim < 0 or dim >= array.ndim:
        raise ValueError(f"Invalid dim: {dim}. Must be between 0 and {array.ndim-1}.")

    return indexes_, dim


def adjust_vector_angle(
    vector_1: Union[NDArray, Tensor, Sequence[float]],
    vector_2: Union[NDArray, Tensor, Sequence[float]],
    scaling_factor: float,
    post_normalize: bool = False,
) -> Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor]]:
    """Move a pair of vectors away or towards each other in the same plane.

    Converge or Diverge a pair of vectors by increasing or decreasing their distance from each other.
    The norm or the length of the vectors is preserved.

    Args:
        vector_1 (NDArray | Tensor): A 1D array of size n representing a word in an n dimensional vector space.
        vector_2 (NDArray | Tensor): A 1D array of size n representing another word in an n dimensional vector space.
        scaling_factor (float): The scaling factor by which the vector difference should be enhanced or diminished. The fraction of distance between the vectors where new vector should land. (sf > 1 diverges the two vectors. sf = 1 leaves the two vectors unchanged. 0.5 < sf < 1 converges the two vectors. sf = 0.5 makes the two vectors equal to their mean. sf = 0 swaps the two vectors. sf < 0.5 move the vectors away from their mean but in opposite direction.)
        post_normalize (bool, optional): Make the magnitude of both output vectors equal to 1 after they have been rotated. Defaults to False.

    Returns:
        Tuple[NDArray | Tensor, NDArray | Tensor]: moved vectors.
    """

    # TODO: make it work for batches of vectors or nd-arrays
    # TODO: Handle NaNs

    if not isinstance(vector_1, (np.ndarray, Tensor)):
        vector_1 = np.array(vector_1)
    if not isinstance(vector_2, (np.ndarray, Tensor)):
        vector_2 = np.array(vector_2)

    v1_norm = ArrayOps.norm(vector_1, dim=None)
    v2_norm = ArrayOps.norm(vector_2, dim=None)

    # make the magnitude of both vectors = 1
    vector_1 = vector_1 / (v1_norm or 1)
    vector_2 = vector_2 / (v2_norm or 1)

    # figure out the dimension of divergence
    v1_minus_v2 = vector_1 - vector_2

    # move each vector away or towards the other by equal amount
    new_v1 = vector_2 + scaling_factor * v1_minus_v2
    new_v2 = vector_1 - scaling_factor * v1_minus_v2

    # make the magnitude of the new vectors = 1
    new_v1 = new_v1 / (ArrayOps.norm(new_v1) or 1)
    new_v2 = new_v2 / (ArrayOps.norm(new_v2) or 1)

    if not post_normalize:
        # restore original magnitudes
        new_v1 = new_v1 * v1_norm
        new_v2 = new_v2 * v2_norm

    # sf > 1 diverges the two vectors
    # new_v1 = v2 + 2.00 * (v1 - v2) = 2 * v1 - v2     # more v1, less v2.
    # new_v2 = v1 - 2.00 * (v1 - v2) = 2 * v2 - v1     # more v2, less v1.

    # sf = 1 leaves the two vectors unchanged
    # new_v1 = v2 + 1.00 * (v1 - v2) = v1
    # new_v2 = v1 - 1.00 * (v1 - v2) = v2

    # 0.5 < sf < 1 converges the two vectors
    # new_v1 = v2 + 0.75 * (v1 - v2) = 0.75 * v1 + 0.25 * v2    # weighted average
    # new_v1 = v1 - 0.75 * (v1 - v2) = 0.75 * v2 + 0.25 * v1    # weighted average

    # sf = 0.5 makes the two vectors equal
    # new_v1 = v2 + 0.50 * (v1 - v2) = 0.5 * v1 + 0.5 * v2   # mean
    # new_v1 = v1 - 0.50 * (v1 - v2) = 0.5 * v2 + 0.5 * v1   # mean

    # sf = 0. swaps the two vectors
    # new_v1 = v2 + 0.00 * (v1 - v2) = v2
    # new_v2 = v1 + 0.00 * (v1 - v2) = v1

    # sf < 0.5 move the vectors away from their mean but in opposite direction
    # new_v1 = v2 + (-1) * (v1 - v2) = 2 * v2 - v1    # more v2, less v1.
    # new_v2 = v1 - (-1) * (v1 - v2) = 2 * v1 - v2    # more v1, less v2.

    return new_v1, new_v2


def align_vectors(
    source_matrix: Union[NDArray, Tensor],
    target_matrix: Union[NDArray, Tensor],
    pre_normalize: bool = True,
) -> Union[NDArray, Tensor]:
    """
    Align the source matrix to the target matrix using the orthogonal transformation.

    Args:
        source_matrix (NDArray | Tensor): A 2D array of shape (dictionary_length, embedding_dimension) containing word vectors from source model (or language).
        target_matrix (NDArray | Tensor): A 2D array of shape (dictionary_length, embedding_dimension) containing word vectors from target model (or language).
        normalize_vectors (bool, optional): Whether to normalize the training vectors before SVD. Defaults to True.

    Returns:
        NDArray | Tensor: An orthogonal transformation which aligns the source language to the target language.

    Note:
        This function supports both NumPy arrays and PyTorch tensors as input.

    """
    # https://github.com/babylonhealth/fastText_multilingual

    if not isinstance(source_matrix, (np.ndarray, Tensor)):
        source_matrix = np.array(source_matrix)
    if not isinstance(target_matrix, (np.ndarray, Tensor)):
        target_matrix = np.array(target_matrix)

    # optionally normalize the training vectors
    if pre_normalize:
        src_norm = ArrayOps.norm(source_matrix, dim=1, keepdim=True)
        src_norm[src_norm == 0] = 1
        source_matrix = source_matrix / src_norm

        tgt_norm = ArrayOps.norm(target_matrix, dim=1, keepdim=True)
        tgt_norm[tgt_norm == 0] = 1
        target_matrix = target_matrix / tgt_norm

    # perform the SVD
    product = source_matrix.T @ target_matrix
    U, _, V = ArrayOps.svd(product)

    # return orthogonal transformation which aligns source language to the target
    return U @ V



from abc import ABC, abstractmethod

import torch


class TextEmbeddingModel(ABC):
    """
    Abstract class for text embedding models.

    Methods:
        embed(text: str) -> torch.Tensor: Embeds text into a vector.
    """

    @abstractmethod
    def embed(self, text: str) -> torch.Tensor:
        """
        Embeds text into a vector.

        Args:
            text (str): Text to embed.

        Returns:
            torch.Tensor: A vector representation of a text.
        """

    # load and save methods
